from pydantic import BaseModel, Field
from typing import List, Literal, Optional, Union


class CACRutherfordIOsettingsLoad(BaseModel):
    """
    Level 3: Class for Input/Output settings for the cable geometry
    """
    load_from_yaml: Optional[bool] = Field(
        default=None,
        description="True to load cable geometry from yaml-file, false to create the geometry.",
    )
    filename: Optional[str] = Field(
        default=None,
        description="Name of the file from which to load the cable geometry.",
    )


class CACRutherfordIOsettingsSave(BaseModel):
    """
    Level 3: Class for Input/Output settings for the cable geometry
    """
    save_to_yaml: Optional[bool] = Field(
        default=None,
        description="True to save cable geometry to yaml-file, false to not save the geometry.",
    )
    filename: Optional[str] = Field(
        default=None,
        description="Name of the file to which to save the cable geometry.",
    )


class CACRutherfordIOsettings(BaseModel):
    """
    Level 3: Class for Input/Output settings for the cable geometry
    """
    load: CACRutherfordIOsettingsLoad = (
        CACRutherfordIOsettingsLoad()
    )
    save: CACRutherfordIOsettingsSave = (
        CACRutherfordIOsettingsSave()
    )

class CACRutherfordExcitationCoils(BaseModel):
    """
    Level 3: Class for Input/Output settings for the cable geometry
    """
    centers: Optional[List[List[float]]] = Field(
        default=None,
        description="List of center points for the centers of the excitations coil regions. Each center point is a list of three elements for x, y, and z (=0) coordinates.",
    )
    widths: Optional[List[float]] = Field(
        default=None,
        description="List of widths of the excitation coil regions.",
    )
    heights: Optional[List[float]] = Field(
        default=None,
        description="List of heights of the excitation coil regions.",
    )

# ============= GEOMETRY ============= #


class CACRutherfordGeometry(BaseModel):
    """
    Level 2: Class for cable geometry parameters
    """
    io_settings: CACRutherfordIOsettings = CACRutherfordIOsettings()

    point_snap_tolerance_relative_to_strand_diameter: Optional[float] = Field(
        default=None,
        description="The maximum distance between two points, relative to the strand diameter, where the points are considered equal (i.e. they 'snap' together).",
    )
    min_roundness_factor: Optional[float] = Field(
        default=None,
        description="Minimum roundness is the ratio between the min -and max radius for the corner circle-arcs.",
    )
    air_radius: Optional[float] = Field(
        default=None, description="Radius of the air region (m)."
    )
    coating_corner_arc_radius: Optional[float] = Field(
        default=0, description="Radius of the corner arcs of the coating (m)."
    )
    coating_thickness: Optional[float] = Field(
        default=0, description="Thickness of the coating (m)."
    )
    keep_strand_area: Optional[bool] = Field(
        default=True, description="If True, the area of the strands are determined by the area of the strand described in 'conductors'. If False, the area of the strands are determined based on the cable geometry inputs."
    )

    excitation_coils: CACRutherfordExcitationCoils = CACRutherfordExcitationCoils()

# ============= MESH ============= #

class CACRutherfordMesh(BaseModel):
    """
    Level 2: Class for FiQuS ConductorAC
    """
    scaling_global: Optional[float] = Field(default=1, description="Global scaling factor for mesh size.")

    strand_mesh_size_ratio: Optional[float] = Field(1, description="Mesh size ratio for the strand, relative to the strand diameter.")
    coating_mesh_size_ratio: Optional[float] = Field(1, description="Mesh size ratio for the coating, relative to the strand diameter.")
    air_boundary_mesh_size_ratio: Optional[float] = Field(1, description="Mesh size ratio for the air boundary, relative to the strand diameter.")

# ============= SOLVE ============= #
# -- General parameters -- #
class CACRutherfordSolveGeneralparameters(BaseModel):
    """
    Level 3: Class for general parameters
    """
    temperature: float = Field(default=1.9, description="Temperature (K) of the strand.")
    superconductor_n_value: Optional[float] = Field(default=30, description="n value for the power law (-), used in current sharing law.")
    superconductor_Ic: Optional[float] = Field(default=350, description="Critical current of the strands (A) (e.g., typical value at T=1.9K and B=10T). Will be taken as a constant as in this model the field dependence is not included"
    " (the main purpose of the model is to verify the more efficient Homogenized Conductor model)."
    " Including field-dependence could be done but is not trivial because is mixes global and local quantities in this Rutherford model with strand discretized individually as stranded conductors.")
    matrix_resistance: Optional[float] = Field(default=6.536208e-04, description="Resistance of the matrix (per unit length) (Ohm/m) for the current sharing law. Kept constant in this model (for simplicity).")

    crossing_coupling_resistance: Optional[float] = Field(default=1e-6, description="Crossing coupling resistance (Ohm).")
    adjacent_coupling_resistance: Optional[float] = Field(default=1e-6, description="Adjacent coupling resistance (Ohm).")

    rho_coating: Optional[float] = Field(default=1e-7, description="Resistivity of coating domain outside of the strands (Ohm.m).")
    rho_strands: Optional[float] = Field(default=1e-12, description="Resistivity of strands, when modelled as massive conductors (Ohm.m).")

    noOfMPITasks: Optional[Union[bool, int]] = Field(
        default=False,
        title="No. of tasks for MPI parallel run of GetDP",
        description=(
            "If integer, GetDP will be run in parallel using MPI. This is only valid"
            " if MPI is installed on the system and an MPI-enabled GetDP is used." 
            " If False, GetDP will be run in serial without invoking mpiexec."
        ),
    )

class CACRutherfordSolveInitialConditions(BaseModel):
    """
    Level 3: Class for initial conditions
    """
    init_from_pos_file: bool = Field(
        default=False, description="Do we initialize the solution at a non-zero field."
    )
    pos_file_to_init_from: Optional[str] = Field(
        default=None,
        description="Name of .pos file for magnetic field (A/m) from which the solution should be initialized."
        " Should be in the Geometry_xxx/Mesh_xxx/ folder in which the Solution_xxx will be saved.",
    )


# -- Source parameters -- #
class CACRutherfordSolveSourceparametersSineSuperimposedDC(BaseModel):
    """
    Level 5: Class for superimposed DC field or current parameters for the sine source
    """
    field_magnitude: Optional[float] = Field(default=0.0, description="DC field magnitude (T) (direction along y-axis). Solution must be initialized with a non-zero field solution stored in a .pos file if non-zero DC field is used.")
    current_magnitude: Optional[float] = Field(default=0.0, description="DC current magnitude (A). Solution must be initialized with a non-zero field solution stored in a .pos file if non-zero DC current is used.")

class CACRutherfordSolveSourceparametersSine(BaseModel):
    """
    Level 4: Class for Sine source parameters
    """
    frequency: Optional[float] = Field(default=None, description="Frequency of the sine source (Hz).")
    field_amplitude: Optional[float] = Field(default=None, description="Amplitude of the sine field (T).")
    current_amplitude: Optional[float] = Field(default=None, description="Amplitude of the sine current (A).")
    field_angle: Optional[float] = Field(default=90, description="Angle of the sine field direction, with respect to the x-axis (degrees).")
    superimposed_DC: CACRutherfordSolveSourceparametersSineSuperimposedDC = CACRutherfordSolveSourceparametersSineSuperimposedDC()

class CACRutherfordSolveSourceparametersPiecewise(BaseModel):
    """
    Level 4: Class for piecewise (linear) source parameters
    """
    source_csv_file: Optional[str] = Field(default=None, description="File name for the from_file source type defining the time evolution of current and field (in-phase). Multipliers are used for each of them. The file should contain two columns: 'time' (s) and 'value' (field/current (T/A)), with these headers. If this field is set, times, applied_fields_relative and transport_currents_relative are ignored.")
    times: Optional[List[float]] = Field(default=None, description="Time instants (s) defining the piecewise linear sources. Used only if source_csv_file is not set. Can be scaled by time_multiplier.")
    applied_fields_relative: Optional[List[float]] = Field(default=None, description="Applied fields relative to multiplier applied_field_multiplier at the time instants 'times'. Used only if source_csv_file is not set.")
    transport_currents_relative: Optional[List[float]] = Field(default=None, description="Transport currents relative to multiplier transport_current_multiplier at the time instants 'times'. Used only if source_csv_file is not set.")
    time_multiplier: Optional[float] = Field(default=None, description="Multiplier for the time values in times (scales the time values). Also used for the time values in the source_csv_file.")
    applied_field_multiplier: Optional[float] = Field(default=None, description="Multiplier for the applied fields in applied_fields_relative. Also used for the values in the source_csv_file.")
    transport_current_multiplier: Optional[float] = Field(default=None, description="Multiplier for the transport currents in transport_currents_relative. Also used for the values in the source_csv_file.")
    field_angle: Optional[float] = Field(default=90, description="Angle of the sine field direction, with respect to the x-axis (degrees).")

class CACRutherfordSolveSourceparametersExcitationCoils(BaseModel):
    """
    Level 4: Class for excitation coils
    """
    enable: Optional[bool] = Field(default=False, description="Are the excitation coils used in the model? (they can exist in the geometry and mesh but be ignored at the solution stage)")
    source_csv_file: Optional[str] = Field(default=None, description="The file should contain a first column with 'time' (s) and one additional column per excitation coil with 'value', which is the TOTAL current (A) per coil (with appropriate sign).")
    

class CACRutherfordSolveSourceparameters(BaseModel):
    """
    Level 3: Class for material properties
    """
    source_type: Literal['sine', 'piecewise'] = Field(
        default='sine',
        description="Time evolution of applied current and magnetic field. Supported options are: sine, sine_with_DC, piecewise_linear, from_list.",
    )
    parallel_resistor: Optional[Union[bool, float]] = Field(
        default=False,
        title="Resistor parallel to the cable",
        description=(
            "If False, no parallel resistor and the current source directly and only feeds the cable."
            " If True, a resistor is placed in parallel with the cable, with a default resistance of 1 Ohm. If float (cannot be zero), this defines the value of the resistance." 
        ),
    )
    boundary_condition_type: str = Field(
        default="Natural",
        description="Boundary condition type. Supported options are: Natural, Essential. Do not use essential boundary condition with induced currents.",
    )
    sine: CACRutherfordSolveSourceparametersSine = CACRutherfordSolveSourceparametersSine()
    piecewise: CACRutherfordSolveSourceparametersPiecewise = CACRutherfordSolveSourceparametersPiecewise()
    excitation_coils: CACRutherfordSolveSourceparametersExcitationCoils = CACRutherfordSolveSourceparametersExcitationCoils()

# -- Numerical parameters -- #
class CACRutherfordSolveNumericalparametersSine(BaseModel):
    """ 
    Level 4: Numerical parameters corresponding to the sine source
    """
    timesteps_per_period: Optional[float] = Field(default=None, description="Initial value for number of time steps (-) per period for the sine source. Determines the initial time step size.")
    number_of_periods_to_simulate: Optional[float] = Field(default=None, description="Number of periods (-) to simulate for the sine source.")

class CACRutherfordSolveNumericalparametersPiecewise(BaseModel):
    """
    Level 4: Numerical parameters corresponding to the piecewise source
    """
    time_to_simulate: Optional[float] = Field(default=None, description="Total time to simulate (s). Used for the piecewise source.")
    timesteps_per_time_to_simulate: Optional[float] = Field(default=None, description="If variable_max_timestep is False. Number of time steps (-) per period for the piecewise source.")
    force_stepping_at_times_piecewise_linear: bool = Field(default=False, description="If True, time-stepping will contain exactly the time instants that are in the times_source_piecewise_linear list (to avoid truncation maximum applied field/current values).")

    variable_max_timestep: bool = Field(default=False, description="If False, the maximum time step is kept constant through the simulation. If True, it varies according to the piecewise definition.")
    times_max_timestep_piecewise_linear: Optional[List[float]] = Field(default=None, description="Time instants (s) defining the piecewise linear maximum time step.")
    max_timestep_piecewise_linear: Optional[List[float]] = Field(default=None, description="Maximum time steps (s) at the times_max_timestep_piecewise_linear. Above the limits, linear extrapolation of the last two values.")

class CACRutherfordSolveNumericalparameters(BaseModel):
    """
    Level 3: Class for numerical parameters
    """
    sine: CACRutherfordSolveNumericalparametersSine = CACRutherfordSolveNumericalparametersSine()
    piecewise: CACRutherfordSolveNumericalparametersPiecewise = CACRutherfordSolveNumericalparametersPiecewise()


class CACRutherfordSolveFormulationparameters(BaseModel):
    """
    Level 3: Class for finite element formulation parameters
    """
    stranded_strands: bool = Field(
        default=True, description="Are the strands solved as 'stranded conductors', i.e., with fixed source current density, and no eddy current effect? Put to True if we solve for homogenized strands."
    )
    rohm: bool = Field(
        default=True, description="Do we use the ROHM model to describe the stranded strand magnetization? This is only relevant with stranded strands, but can be used without (without much meaning). If fase, solves with permeability mu0."
    )
    rohf: bool = Field(
        default=True, description="Do we use the ROHF model to describe the stranded strand voltage and inductance? This is only possible with stranded strands. If stranded_strands=false, rohf is considered false as well."
    )

class CACRutherfordSolveFrequencydomainsolverFrequencysweep(BaseModel):
    """
    Level 4: Class for frequency sweep settings
    """
    run_sweep: bool = Field(
        default=False, description="Set True to run a frequency sweep (logarithmic)."
    )
    start_frequency: float = Field(
        default=1, description="Start frequency (Hz) of the sweep."
    )
    end_frequency: float = Field(
        default=100, description="End frequency (Hz) of the sweep."
    )
    number_of_frequencies: int = Field(
        default=10, description="Number of frequencies in the sweep."
    )


    
class CACRutherfordSolveFrequencydomainsolver(BaseModel):
    """
    Level 3: Class for frequency domain solver settings
    """
    enable: bool = Field(
        default=False, description="Set True to enable the frequency domain solver."
    )
    frequency_sweep: CACRutherfordSolveFrequencydomainsolverFrequencysweep = (
        CACRutherfordSolveFrequencydomainsolverFrequencysweep()
    )
    


class CACRutherfordSolve(BaseModel):
    """
    Level 2: Class for FiQuS ConductorAC
    """
    pro_template: Optional[str] = Field(
        default=None, description="Name of the .pro template file."
    )
    conductor_name: Optional[str] = Field(
        default=None, description="Name of the conductor."
    )
    formulation_parameters: CACRutherfordSolveFormulationparameters = (
        CACRutherfordSolveFormulationparameters()
    )
    general_parameters: CACRutherfordSolveGeneralparameters = (
        CACRutherfordSolveGeneralparameters()
    )
    initial_conditions: CACRutherfordSolveInitialConditions = (
        CACRutherfordSolveInitialConditions()
    )
    frequency_domain_solver: CACRutherfordSolveFrequencydomainsolver = (
        CACRutherfordSolveFrequencydomainsolver()
    )
    source_parameters: CACRutherfordSolveSourceparameters = (
        CACRutherfordSolveSourceparameters()
    )
    numerical_parameters: CACRutherfordSolveNumericalparameters = (
        CACRutherfordSolveNumericalparameters()
    )


# ============= POSTPROC ============= #
class CACRutherfordPostprocBatchpostprocLossMapCrossSection(BaseModel):
    """
    Level 5: Class with settings for plotting a cross-section of the loss map.
    """
    plot_cross_section: bool = Field(
        default=False, description="Set True to plot a cross-section of the loss map."
    )
    save_plot: bool = Field(default=False, description="Set True to save the plot.")
    filename: str = Field(default="cross_section", description="Name of the plot file.")
    axis_to_cut: str = Field(
        default="x", description="Axis to cut for the cross-section."
    )
    cut_value: float = Field(
        default=0, description="Value of the axis to cut for the cross-section."
    )

    ylabel: str = Field(default="Loss", description="Label of the y-axis.")
    title: Optional[str] = Field(
        default=None,
        description="Title of the plot. The placeholder <<cut_value>> can be used to indicate the value of the cut axis.",
    )


class CACRutherfordPostprocBatchpostprocLossMapCrossSectionSweep(BaseModel):
    """
    Level 5: Class with settings for animating a cross-section sweep of the loss map along one axis.
    """
    animate_cross_section_sweep: bool = Field(
        default=False,
        description="Set True to animate a cross-section sweep of the loss map along one axis.",
    )
    save_plot: bool = Field(
        default=False, description="Set True to save the animation."
    )
    filename: str = Field(
        default="crossSectionSweep", description="Name of the animation file."
    )
    axis_to_sweep: str = Field(
        default="x", description="Axis to sweep for the animation."
    )
    ylabel: str = Field(default="Loss", description="Label of the y-axis.")
    title: Optional[str] = Field(
        default=None,
        description="Title of the plot. Use <<sweep_value>> to indicate the value of the sweep axis.",
    )


class CACRutherfordPostprocBatchpostprocLossMap(BaseModel):
    """
    Level 4: Class with settings for generating loss maps
    """
    produce_loss_map: bool = Field(
        default=False, description="Set True to produce a loss map."
    )
    save_plot: bool = Field(default=False, description="Set True to save the plot.")
    filename: str = Field(default="loss_map", description="Name of the plot file.")
    x_val: Optional[str] = Field(
        default=None, description="Parameter to be plotted on the x-axis. This field corresponds to a parameter in the input YAML-file. E.g. 'solve.source_parameters.sine.frequency' will plot the loss map for different frequencies."
    )
    y_val: Optional[str] = Field(
        default=None, description="Parameter to be plotted on the y-axis. This field corresponds to a parameter in the input YAML-file. E.g. 'solve.source_parameters.sine.field_amplitude' will plot the loss map for different applied field amplitudes."
    )
    x_steps: int = Field(default=20, description="Number of steps on the x-axis.")
    y_steps: int = Field(default=20, description="Number of steps on the y-axis.")
    loss_type: Literal['TotalLoss', 'FilamentLoss', 'CouplingLoss', 'EddyLoss'] = Field(
        default='TotalLoss',
        description="Type of loss to be plotted. Supported options are: TotalLoss, FilamentLoss, CouplingLoss, EddyLoss."
    )
    x_log: bool = Field(
        default=True, description="Set True to plot x-axis in log-scale."
    )
    y_log: bool = Field(
        default=True, description="Set True to plot y-axis in log-scale."
    )
    loss_log: bool = Field(
        default=True, description="Set True to plot loss in log-scale."
    )
    x_norm: float = Field(default=1, description="Normalization factor for x-axis.")
    y_norm: float = Field(default=1, description="Normalization factor for y-axis.")
    loss_norm: float = Field(default=1, description="Normalization factor for the AC-loss.")
    show_datapoints: bool = Field(
        default=True, description="Set True to show markers for all the datapoints in the loss map."
    )

    title: Optional[str] = Field(default=None, description="Title for the plot.")
    xlabel: Optional[str] = Field(default=None, description="Label for the x-axis.")
    ylabel: Optional[str] = Field(default=None, description="Label for the y-axis.")

    # lossType_dominance_contour: CACStrandPostprocBatchpostprocLossMapDominanceCountour = (
    #     CACStrandPostprocBatchpostprocLossMapDominanceCountour()
    # )

    show_loss_type_dominance_contour: bool = Field(
        default=False,
        description="Set True to plot a contour curve separating regions where different loss types dominate. ",
    )

    cross_section: CACRutherfordPostprocBatchpostprocLossMapCrossSection = (
        CACRutherfordPostprocBatchpostprocLossMapCrossSection()
    )
    cross_section_sweep: CACRutherfordPostprocBatchpostprocLossMapCrossSectionSweep = (
        CACRutherfordPostprocBatchpostprocLossMapCrossSectionSweep()
    )


class CACRutherfordPostprocBatchpostprocPlot2d(BaseModel):
    """
    Level 4: Class for 2D plot settings
    """
    produce_plot2d: bool = Field(
        default=False, description="Set True to produce a 2D plot."
    )
    combined_plot: bool = Field(
        default=False,
        description="Set True to produce a combined plot for all simulations. If False, a separate plot is produced for each simulation.",
    )
    save_plot: bool = Field(default=False, description="Set True to save the plot.")
    filename: str = Field(default="plot2d", description="Name of the plot file.")
    x_val: Optional[str] = Field(
        default=None, description="Value to be plotted on the x-axis. Parameters in the input YAML-file and class-variables from the plotter 'SimulationData' class can be accessed trough the notation << . >>. E.g. '<<solve.source_parameters.sine.frequency>>' will create a 2D plot with frequency on the x-axis. '<<time>>' will create a plot with time on the x-axis."
    )
    y_vals: Optional[List[str]] = Field(
        default=None, description=" List of values to be plotted on the y-axis. Parameters in the input YAML-file and class-variables from the plotter 'SimulationData' class can be accessed trough the notation << . >>. E.g. total AC-loss per cycle can be accessed as ['<<total_power_per_cycle['TotalLoss_dyn']>>']."
    )
    labels: Optional[List[str]] = Field(
        default=None,
        description="List of labels for the plot. Each label corresponding to a value in y_val.",
    )
    linestyle: Optional[str] = Field(
        default=None, description="Linestyle for the plot."
    )

    title: Optional[str] = Field(default=None, description="Title for the plot.")
    xlabel: Optional[str] = Field(default=None, description="Label for the x-axis.")
    ylabel: Optional[str] = Field(default=None, description="Label for the y-axis.")
    x_log: bool = Field(default=False, description="Set True to plot x-axis in log-scale.")
    y_log: bool = Field(default=False, description="Set True to plot y-axis in log-scale.")
    legend: bool = Field(default=True, description="Set True to show legend.")


class CACRutherfordPostprocBatchpostprocFilter(BaseModel):
    """
    Level 4: Field for filtering simulations based on simulation parameters for batch post-processing
    """
    apply_filter: bool = Field(
        default=False,
        description="Set True to filter simulations by parameters from the input YAML-file.",
    )
    filter_criterion: Optional[str] = Field(
        default=None,
        description="Criterion used to filter simulations based on simulation parameters. For example will '<<solve.source_parameters.sine.frequency>> > 100' disregard simulations done with frequencies lower than 100Hz.",
    )


class CACRutherfordPostprocBatchpostprocSort(BaseModel):
    """
    Level 4: Field for sorting simulations based on simulation parameters for batch post-processing
    """
    apply_sort: bool = Field(default=False, description="Set True to sort simulations.")
    sort_key: Optional[str] = Field(
        default=None,
        description="Criterion used to sort simulations based on simulation parameters. For example will 'sd.total_power_per_cycle['TotalLoss'] sort simulations based on the total loss.",
    )


class CACRutherfordPostprocBatchpostproc(BaseModel):
    """
    Level 3: Class for batch post-processing settings
    """
    postProc_csv: Optional[str] = Field(
        default=None,
        description="Name of the .csv file for post-processing (without file extension). This file specifies the simulations to be post-processed. The file is structured into three columns, specifying the folder names to access the simulation results: 'input.run.geometry', 'input.run.mesh' and 'input.run.solve'. Each row corresponds to a simulation to be post-processed.",
    )
    output_folder: Optional[str] = Field(
        default=None,
        description="Batch post-processing creates a folder with the given name in the output directory, where all the plots are saved.",
    )
    filter: CACRutherfordPostprocBatchpostprocFilter = CACRutherfordPostprocBatchpostprocFilter()
    sort: CACRutherfordPostprocBatchpostprocSort = CACRutherfordPostprocBatchpostprocSort()
    loss_map: CACRutherfordPostprocBatchpostprocLossMap = CACRutherfordPostprocBatchpostprocLossMap()
    plot2d: CACRutherfordPostprocBatchpostprocPlot2d = CACRutherfordPostprocBatchpostprocPlot2d()


class CACRutherfordPostprocPlotInstPower(BaseModel):
    """
    Level 3: Class with settings for generating plots of instantaneous power
    """
    show: bool = Field(default=False, description="Creates a plot for the calculated instantaneous AC loss (W/m) as a function of time (s).")
    title: str = Field(default="Instantaneous Power", description="Title for the plot.")
    save: bool = Field(default=False, description="Set True to save the plot.")
    save_file_name: str = Field(
        default="instantaneous_power", description="Name of the plot file."
    )


class CACRutherfordPostprocCleanup(BaseModel):
    """
    Level 3: Class for cleanup settings
    """
    remove_pre_file: bool = Field(
        default=False,
        description="Set True to remove the .pre-file after post-processing, to save disk space.",
    )
    remove_res_file: bool = Field(
        default=False,
        description="Set True to remove the .res-file after post-processing, to save disk space.",
    )
    remove_msh_file: bool = Field(
        default=False,
        description="Set True to remove the .msh-file after post-processing, to save disk space.",
    )


class CACRutherfordPostproc(BaseModel):
    """
    Level 2: Class for FiQuS ConductorAC
    """
    generate_pos_files: bool = Field(
        default=True,
        description="Set True to generate .pos-files during post-processing",
    )
    plot_instantaneous_power: CACRutherfordPostprocPlotInstPower = (
        CACRutherfordPostprocPlotInstPower()
    )
    compute_current_per_filament: bool = Field(
        default=False,
        description="Computes current in every filament, with decomposition into magnetization and transport current.",
    )
    save_last_current_density: Optional[str] = Field(
        default=None,
        description="Saves the last current density field solution (out-of-plane) in the file given as a string."
        " The '.pos' extension will be appended to it. Nothing is done if None."
        " This can be for using the current density as an initial condition (but not implemented yet).",
    )
    save_last_magnetic_field: Optional[str] = Field(
        default=None,
        description="Saves the last magnetic field solution (in-plane) in the file given as a string."
        " The '.pos' extension will be appended to it. Nothing is done if None."
        " This is for using the magnetic field as an initial condition for another resolution.",
    )
    cleanup: CACRutherfordPostprocCleanup = CACRutherfordPostprocCleanup()
    batch_postproc: CACRutherfordPostprocBatchpostproc = CACRutherfordPostprocBatchpostproc()


# ============= BASE ============= #
class CACRutherford(BaseModel):
    """
    Level 1: Class for FiQuS ConductorAC
    """
    type: Literal["CACRutherford"]
    geometry: CACRutherfordGeometry = CACRutherfordGeometry()
    mesh: CACRutherfordMesh = CACRutherfordMesh()
    solve: CACRutherfordSolve = CACRutherfordSolve()
    postproc: CACRutherfordPostproc = CACRutherfordPostproc()



from pydantic import BaseModel, Field

from typing import List, Union, Optional, Literal


############################
# Circuit
class Circuit_Class(BaseModel):
    """
        Level 1: Class for the circuit parameters
    """
    field_circuit: Optional[bool] = Field(
        default = False,
        description = "Allows to use Field-Circuit Coupling equations in the model.")
    R_circuit: Optional[float] = None             # R_circuit
    L_circuit: Optional[float] = None             # Lcir (SIGMA)
    R_parallel: Optional[float] = None

 
############################
# Power Supply (aka Power Converter)
class PowerSupplyClass(BaseModel):
    """
        Level 1: Class for the power supply (aka power converter)
    """
    I_initial: Optional[float] = Field(
        default=None,
        description="Initial current in the magnet. Propagated differently in various tools and obsolete # I00 (LEDET), I_0 (SIGMA), I0 (BBQ)")
    t_off: Optional[float] = Field(
        default=None,
        description="Time of switching off the switch next to current controlled source. t_PC (LEDET)")
    t_control_LUT: List[float] = Field(
        default=[],
        description="List of time values [s] for linear piece wise time function of current controlled source. t_PC_LUT (LEDET)")
    I_control_LUT: List[float] = Field(
        default=[],
        description="List of current values [A] for linear piece wise time function of current controlled source. I_PC_LUT (LEDET)")
    R_crowbar: Optional[float] = Field(
        default=0,
        description="Crowbar resistance in forward direction [Ohm]. Rcrow (SIGMA), RCrowbar (ProteCCT)")
    L_crowbar: Optional[float] = Field(
        default=0,
        description="Crowbar inductance in forward direction [H].")
    Ud_crowbar: Optional[float] = Field(
        default=0,
        description="Crowbar diode voltage in forward direction [V].")
    R_c_r: Optional[float] = Field(
        default=0,
        description="Crowbar resistance in reverse direction [Ohm].")
    L_c_r: Optional[float] = Field(
        default=0,
        description="Crowbar inductance in reverse direction [H].")
    Ud_c_r: Optional[float] = Field(
        default=0,
        description="Crowbar diode voltage in reverse direction [V].")
    R_1: Optional[float] = Field(
        default=0,
        description="Resistance R1 [Ohm].")
    L_1: Optional[float] = Field(
        default=0,
        description="Inductance L1 [H].")
    R_2: Optional[float] = Field(
        default=0,
        description="Resistance R2 [Ohm].")
    L_2: Optional[float] = Field(
        default=0,
        description="Inductance L2 [H].")
    C: Optional[float] = Field(
        default=0,
        description="Capacitance C [F].")
    R_3: Optional[float] = Field(
        default=0,
        description="Resistance R3 [Ohm].")
    L_3: Optional[float] = Field(
        default=0,
        description="Inductance L3 [H].")

############################
# Quench Protection
class EnergyExtraction(BaseModel):
    """
        Level 2: Class for the energy extraction parameters
    """
    t_trigger: Optional[float] = Field(
        default=99999.0,
        description="Trigger time on the positive lead [s]. tEE (LEDET), tSwitchDelay (ProteCCT)")
    R_EE: Optional[float] = Field(
        default=0,
        description="Energy extraction resistance on the positive lead [Ohm]. R_EE_triggered (ProteCCT)")
    power_R_EE: Optional[float] = Field(
        default=0.0,
        description="Varistor power component, R(I) = R_EE*abs(I)^power_R_EE on the positive lead [-]. RDumpPower (ProteCCT)")
    L: Optional[float] = Field(
        default=0,
        description="Inductance in series with resistor on the positive lead [H].")
    C: Optional[float] = Field(
        default=0,
        description="Snubber capacitance in parallel to the EE switch on the positive lead [F].")
    L_c: Optional[float] = Field(
        default=0,
        description="Inductance in the snubber capacitance branch in parallel to the EE switch on the positive lead [H].")
    R_c: Optional[float] = Field(
        default=0,
        description="Resistance in the snubber capacitance branch in parallel to the EE switch on the positive lead [Ohm].")
    Ud_snubber: Optional[float] = Field(
        default=0,
        description="Forward voltage of diode in the snubber capacitance branch in parallel to the EE switch on the positive lead [V].")
    L_s: Optional[float] = Field(
        default=0,
        description="Inductance in the EE switch branch on the positive lead [H].")
    R_s: Optional[float] = Field(
        default=0,
        description="Resistance in the EE switch branch on the positive lead [Ohm].")
    Ud_switch: Optional[float] = Field(
        default=0,
        description="Forward voltage of diode in the EE switch branch on the positive lead [V].")

    t_trigger_n: Optional[float] = Field(
        default=99999.0,
        description="Trigger time on the negative lead [s]. tEE (LEDET), tSwitchDelay (ProteCCT)")
    R_EE_n: Optional[float] = Field(
        default=0,
        description="Energy extraction resistance on the negative lead [Ohm]. R_EE_triggered (ProteCCT)")
    power_R_EE_n: Optional[float] = Field(
        default=0.0,
        description="Varistor power component, R(I) = R_EE*abs(I)^power_R_EE on the negative lead [-]. RDumpPower (ProteCCT)")
    L_n: Optional[float] = Field(
        default=0,
        description="Inductance in series with resistor on the negative lead [H].")
    C_n: Optional[float] = Field(
        default=0,
        description="Snubber capacitance in parallel to the EE switch on the negative lead [F].")
    L_c_n: Optional[float] = Field(
        default=0,
        description="Inductance in the snubber capacitance branch in parallel to the EE switch on the negative lead [H].")
    R_c_n: Optional[float] = Field(
        default=0,
        description="Resistance in the snubber capacitance branch in parallel to the EE switch on the negative lead [Ohm].")
    Ud_snubber_n: Optional[float] = Field(
        default=0,
        description="Forward voltage of diode in the snubber capacitance branch in parallel to the EE switch on the negative lead [V].")
    L_s_n: Optional[float] = Field(
        default=0,
        description="Inductance in the EE switch branch on the negative lead [H].")
    R_s_n: Optional[float] = Field(
        default=0,
        description="Resistance in the EE switch branch on the negative lead [Ohm].")
    Ud_switch_n: Optional[float] = Field(
        default=0,
        description="Forward voltage of diode in the EE switch branch on the negative lead [V].")


class QuenchHeater(BaseModel):
    """
        Level 2: Class for the quench heater parameters
    """
    N_strips: Optional[int] = Field(
        default=None,
        description="Number of quench heater traces (typically 2 traces make one pad)")
    t_trigger: List[float] = Field(
        default=[],
        description="Trigger times list of of quench heaters [s]")
    U0: List[float] = Field(
        default=[],
        description="Initial charging voltages list of capacitor for the trance (not full pad!) [V]")
    C: List[float] = Field(
        default=[],
        description="Capacitances list of quench heater power supply for the trance (not full pad!) [H]")
    R_warm: List[float] = Field(
        default=[],
        description="Internal resistances list of quench heater power supply and/or additional resistance added to limit the heater current for the trance (not full pad!) [Ohm]")
    w: List[float] = Field(
        default=[],
        description="Widths list of quench heater trance stainless steel part [m]")
    h: List[float] = Field(
        default=[],
        description="Thickness list of quench heater trance stainless steel part [m]")
    s_ins: Union[List[float], List[List[float]]] = Field(
        default=[],
        description="Thickness list of quench heater insulation between the stainless steel part and conductor insulation [m]"
                    "This could be a list of list to specify multiple material thicknesses")
    type_ins: Union[List[str], List[List[str]]] = Field(
        default=[],
        description="Material names list of quench heater insulation between the stainless steel part and conductor insulation [-]"
                    "This could be a list of list to specify multiple material names")
    s_ins_He: Union[List[float], List[List[float]]] = Field(
        default=[],
        description="Material names list of quench heater insulation between the stainless steel part and helium bath [-]"
                    "This could be a list of list to specify multiple material thicknesses")
    type_ins_He: Union[List[str], List[List[str]]] = Field(
        default=[],
        description="Material names list of quench heater insulation between the stainless steel part and helium bath [-]"
                    "This could be a list of list to specify multiple material names")
    l: List[float] = Field(
        default=[],
        description="Lengths list of quench heaters [m]. Typically equal to magnet length.")
    l_copper: List[float] = Field(
        default=[],
        description="Lengths list of copper laminations of quench heaters [m].")
    l_stainless_steel: List[float] = Field(
        default=[],
        description="Lengths list of stainless steel only sections of quench heaters [m].")
    f_cover: List[float] = Field(
        default=[],
        description="List of fraction of stainless steel cover. This is l_stainless_steel/(l_stainless_steel+l_copper). Marked as obsolete, but still specified in some models [-].")
    iQH_toHalfTurn_From: List[int] = Field(
        default=[],
        description="List of heater numbers (1 based) equal to the length of turns that are covered by (i.e. thermally connected to) quench heaters.")
    iQH_toHalfTurn_To: List[int] = Field(
        default=[],
        description="List of turn numbers (1 based) that are covered by (i.e. thermally connected to) quench heaters.")
    turns_sides: List[str] = Field(
        default=[],
        description="List of letters specifying side of turn where quench heater is placed. Only used in FiQuS Multipole module."
                    "Possible sides are: 'o' - outer, 'i' - inner, 'l' - lower angle, 'h' - higher angle.")
    quench_propagation: Optional[Literal['2Dx1D','full']] = Field(
        default = "full",
        description = 'Enables to have a variable length for the quench heater, different from the full magnet length.')
    quench_prop_model: Optional[Literal['Wilson','Ts']] = Field(
        default = "Ts",
        description = 'Selects the model used for the material properties of the quench propagation. "Wilson" uses a scaled cv and Ts uses the cv at Ts.')
    NZPV_multiplier: Optional[float] = Field(
        default = 1.0,
        description = 'Factor that multiplies the Normal Zone Propagation Velocity')
    L_QH_offset: Optional[List[float]] = Field(
        default = [],
        description= 'Offset of the quench heater strip from the referrence point located  at the middle of the magnet length. Positive values move the quench heater towards higher z values (move quench heater strip towards the front ofthe magnet).')

class CLIQ_Class(BaseModel):
    """
        Level 2: Class for the CLIQ parameters
    """
    t_trigger: Optional[float] = Field(
        default=99999.0,
        description="Trigger time of CLIQ unit [s].")
    current_direction: List[int] = Field(
        default=[],
        description="Polarity of current in groups specified as a list with length equal to the number of groups [-].")
    sym_factor: Optional[float] = Field(
        default=None,
        description="Obsolete.")
    N_units: Optional[int] = Field(
        default=None,
        description="Obsolete.")
    U0: Optional[float] = Field(
        default=0,
        description="Initial charging voltage of CLIQ unit [V].")
    C: Optional[float] = Field(
        default=0,
        description="Capacitance of CLIQ unit [F].")
    R: Optional[float] = Field(
        default=0,
        description="Resistance of CLIQ unit [Ohm].")
    L: Optional[float] = Field(
        default=0,
        description="Inductance of CLIQ unit [H].")
    I0: Optional[float] = Field(
        default=None,
        description="Obsolete.")


class ESC_Class(BaseModel):
    """
        Level 2: Class for the ESC parameters
    """
    t_trigger: List[float] = Field(
        default=[],
        description="Trigger time of ESC units [s] given as a list with length corresponding to the number of ESC units.")
    U0: List[float] = Field(
        default=[],
        description="Initial charging voltage of ESC units [V] given as a list with length corresponding to the number of ESC units."
                    "The unit is grounded in the middle, so the voltage to ground is half of this value")
    C: List[float] = Field(
        default=[],
        description="Capacitance of ESC units [F] given as a list with length corresponding to the number of ESC units."
                    "The unit is grounded in the middle, with two capacitors in series with value of 2C")
    L: List[float] = Field(
        default=[],
        description="Parasitic inductance of ESC units [H] given as a list with length corresponding to the number of ESC units."
                    "The unit is grounded in the middle, with two capacitors in series with value of 2C")
    R_unit: List[float] = Field(
        default=[],
        description="Internal resistance of ESC units [Ohm] given as a list with length corresponding to the number of ESC units.")
    R_leads: List[float] = Field(
        default=[],
        description="Resistance of leads from ESC coil to ESC diode connections [Ohm] given as a list with length corresponding to the number of ESC units.")
    Ud_Diode: List[float] = Field(
        default=[],
        description="Forward diodes voltage across ESC coils [V] given as a list with length corresponding to the number of ESC units.")
    L_Diode: List[float] = Field(
        default=[],
        description="Inductance in series with diodes across ESC coils [V] given as a list with length corresponding to the number of ESC units.")


class FQPCs_Class(BaseModel):
    """
        Level 2: Class for the FQPLs parameters for protection
    """
    enabled: Optional[List[bool]] = Field(
        default=None,
        description="List of booleans specifying which FQPC is enabled.")
    names: Optional[List[str]] = Field(
        default=None,
        description="List of names to use in gmsh and getdp. Any unique ASCII strings would work")
    fndpls: Optional[List[int]] = Field(
        default=None,
        description="List of FQPC number of divisions per length at geometry level [-]")
    fwws: Optional[List[float]] = Field(
        default=None,
        description="List of FQPC wire widths (assuming rectangular). For theta = 0 this is x dimension. Works at geometry level [-]")
    fwhs: Optional[List[float]] = Field(
        default=None,
        description="List of FQPC wire heights (assuming rectangular). For theta = 0 this is y dimension. Works at geometry level [-]")
    r_ins: Optional[List[float]] = Field(
        default=None,
        description="List of FQPC inner diameter of (e.g. for CCT magnet). For theta = 0 this is x dimension. Works at geometry level [-]")
    r_bs: Optional[List[float]] = Field(
        default=None,
        description="List of FQPC radiuses for bending the fqpl by 180 degrees at the end of the magnet to go backward. Works at geometry level [-]")
    n_sbs: Optional[List[int]] = Field(
        default=None,
        description="List of FQPC number of 'bending segments' for the 180 degrees turn. Works at geometry level [-]")
    thetas: Optional[List[float]] = Field(
        default=None,
        description="List of FQPC rotation in deg from x+ axis towards y+ axis about z axis. Works at geometry level [-]")
    z_starts: Optional[List[str]] = Field(
        default=None,
        description="List of z coordinates for the air boundary to start at. These are string with either: z_min or z_max key from the Air region. Works at geometry level [-]")
    z_ends: Optional[List[float]] = Field(
        default=None,
        description="List of z coordinates for the air boundary to end at. These are string with either: z_min or z_max key from the Air region. Works at geometry level [-]")
    currents: Optional[List[float]] = Field(
        default=None,
        description="List of FQPC currents for a magnetostatic solution. Works at solve level [-]")
    sigmas: Optional[List[float]] = Field(
        default=None,
        description="List of FQPC electrical conductivity for a magnetostatic solution. Works at solve level [-]")
    mu_rs: Optional[List[float]] = Field(
        default=None,
        description="List of FQPC magnetic permeability  for a magnetostatic solution. Works at solve level [-]")
    th_conns_def: Optional[List[List]] = Field(
        default=None,
        description="List of lists specifying thermal connections for LEDET to connect the FQPCs to the other turns of the magnet")

class SourceSine(BaseModel):
    """
    Level 3: Class for Sine source parameters for E-CLIQ
    """
    frequency: Optional[float] = Field(default=None, description="Frequency of the sine source [Hz].")
    current_amplitude: Optional[float] = Field(default=None, description="Amplitude of the sine current (A/turn).")
    number_of_periods: Optional[float] = Field(default=None, description="Number of periods of ECLIQ power supply [-].")
    number_of_turns: Optional[int] = Field(default=None, description="Number of turns that conform ECLIQ [-].")

class SourcePiecewise(BaseModel):
    """
    Level 3 Class for piecewise (linear) source parameters for E-CLIQ
    """
    csv_file: Optional[str] = Field(default=None,
                                    description="File name for the from_file source type defining the time evolution of current. Multipliers are used for each of them. The file should contain two columns: 'time' (s) and 'current' (A), with these headers. If this field is set, times and currents are ignored.")
    times: Optional[List[float]] = Field(default=None,
                                         description="Time instants (s) defining the piecewise linear sources. Used only if source_csv_file is not set. Can be scaled by time_multiplier.")
    currents: Optional[List[float]] = Field(default=None,
                                            description="E-CLIQ coil currents relative to current_multiplier at the time instants 'times'. Used only if source_csv_file is not set.")
    time_multiplier: Optional[float] = Field(default=None,
                                             description="Multiplier for the time values in times (scales the time values). Also used for the time values in the source_csv_file.")
    current_multiplier: Optional[float] = Field(default=None,
                                                description="Multiplier for the E-CLIQ coil currents in currents. Also used for the values in the source_csv_file.")


class E_CLIQ_Class(BaseModel):
    """
        Level 2: Class for the E-CLIQ parameters for protection
    """
    t_trigger: Optional[List[float]] = Field(
        default=[],
        description="Trigger time of E-CLIQ current sources [s] given as a list with length corresponding to the number of E-CLIQ units.")
    R_leads: Optional[List[float]] = Field(
        default=None,
        description="List of E-CLIQ unit lead resistances [Ohm]. List length corresponding to the number of E-CLIQ units.")
    L_leads: Optional[List[float]] = Field(
        default=None,
        description="List of E-CLIQ unit lead inductances [H]. List length corresponding to the number of E-CLIQ units.")
    source_type: Literal['sine', 'piecewise', None] = Field(
        default=None,
        description="Time evolution of applied current. Supported options are: sine, piecewise.")
    sine: SourceSine = Field(
        default=SourceSine(),
        description="Definition of sine current source parameters.")
    piecewise: SourcePiecewise = Field(
        default=SourcePiecewise(),
        description="Definition of piecewise current source parameters.")
    quench_propagation: Optional[Literal['2Dx1D','full']] = Field(
        default = "full",
        description = 'Enables to have a variable length for the ecliq implementation, different from the full magnet length. It only affects the Thermal Behaviour of the model.')
    quench_prop_model: Optional[Literal['Wilson','Ts']] = Field(
        default = "Ts",
        description = 'Selects the model used for the material properties of the quench propagation. "Wilson" uses a scaled cv with a function of T_bath and Ts and Ts uses the cv at Ts.')
    NZPV_multiplier: Optional[float] = Field(
        default = 1.0,
        description = 'Factor that multiplies the Normal Zone Propagation Velocity')
    N_units: Optional[List[int]] = Field(
        default = [],
        description= 'Number of E-CLIQ units along the magnet length per ecliq coil. It must be an odd number for symmetry reasons.')
    h_ecliq: Optional[List[float]] = Field(
        default = [],
        description = 'Spacing between the ecliq coils along the magnet length (m).')
    l_ecliq: Optional[List[float]] = Field(
        default = [],
        description = 'length of the ecliq coils along the magnet length (m).')
    L_ecliq_offset: Optional[List[float]] = Field(
        default=[],
        description='Offset of the quench heater strip from the referrence point located  at the middle of the magnet length. Positive values move the quench heater towards higher z values (move quench heater strip towards the front ofthe magnet).')
    iECLIQ_toHalfTurn_From: Optional[List[int]] = Field(
        default=[],
        description='List of coils to which the ECLIQ units are connected from, to which half turns they are in direct contact with.'
    )
    iECLIQ_toHalfTurn_To: Optional[List[int]] = Field(
        default=[],
        description='List of half turns to whom the ECLIQ Units are in direct contact with.'
    )

class QuenchDetection(BaseModel):
    """
    Level 2: Class for FiQuS
    """

    voltage_thresholds: Optional[List[float]] = Field(
        default=None,
        title="List of quench detection voltage thresholds",
        description="Voltage thresholds for quench detection. The quench detection will be triggered when the voltage exceeds these thresholds continuously for a time larger than the discrimination time.",
    )

    discrimination_times: Optional[List[float]] = Field(
        default=None,
        title="List of quench detection discrimination times",
        description="Discrimination times for quench detection. The quench detection will be triggered when the voltage exceeds the thresholds continuously for a time larger than these discrimination times.",
    )

    voltage_tap_pairs: Optional[List[List[int]]] = Field(
        default=None,
        title="List of quench detection voltage tap pairs",
        description="Voltage tap pairs for quench detection. The voltage difference between these pairs will be used for quench detection.",
    )

class QuenchProtection(BaseModel):
    """
        Level 1: Class for quench protection
    """
    Energy_Extraction: EnergyExtraction = EnergyExtraction()
    Quench_Heaters: QuenchHeater = QuenchHeater()
    CLIQ: CLIQ_Class = CLIQ_Class()
    ESC: ESC_Class = ESC_Class()
    FQPCs: FQPCs_Class = FQPCs_Class()
    E_CLIQ: E_CLIQ_Class = E_CLIQ_Class()



import unittest
from tests.utils.fiqus_test_classes import FiQuSMeshTests


class TestMeshGenerators(FiQuSMeshTests):
    def test_Pancake3D(self):
        """
        Checks if Pancake3D geometry generators work correctly by comparing the number
        of entities in the generated geometry file to the reference file that was
        checked manually.
        """
        model_names = [
            "TEST_Pancake3D_REF",
            "TEST_Pancake3D_REFStructured",
            "TEST_Pancake3D_TSA",
            "TEST_Pancake3D_TSAStructured",
            "TEST_Pancake3D_TSAInsulating",
            "TEST_Pancake3D_TSAInsulatingJcVsLength"
        ]

        for model_name in model_names:
            with self.subTest(model_name=model_name):
                data_model = self.get_data_model(model_name)

                # data_model can be modified here if necessary
                # Example:

                # data_model.magnet.mesh.wi.axne = 30

                self.generate_mesh(data_model, model_name)

                # Compare the number of entities with the reference file:
                mesh_file = self.get_path_to_generated_file(
                    data_model=data_model, model_name=model_name, file_extension="msh"
                )
                reference_file = self.get_path_to_reference_file(
                    data_model=data_model, model_name=model_name, file_extension="msh"
                )
                self.compare_mesh_qualities(mesh_file, reference_file)

                # Compare the regions files:
                regions_file = self.get_path_to_generated_file(
                    data_model=data_model,
                    model_name=model_name,
                    file_extension="regions",
                )
                reference_regions_file = self.get_path_to_reference_file(
                    data_model=data_model,
                    model_name=model_name,
                    file_extension="regions",
                )
                self.compare_json_or_yaml_files(regions_file, reference_regions_file)

    def test_ConductorAC_Strand(self):
        """
        Checks if ConductorAC geometry generators work correctly by comparing the number
        of entities in the generated geometry file to the reference file that was
        checked manually.
        """
        model_names = [
            "TEST_CAC_Strand_adaptiveMesh",
            "TEST_CAC_Strand_hexFilaments",
            "TEST_CAC_wireInChannel",
        ]

        for model_name in model_names:
            with self.subTest(model_name=model_name):
                data_model = self.get_data_model(model_name)

                # data_model can be modified here if necessary
                # Example:

                # data_model.magnet.mesh.wi.axne = 30

                self.generate_mesh(data_model, model_name)

                # Compare the number of entities with the reference file:
                mesh_file = self.get_path_to_generated_file(
                    data_model=data_model, model_name=model_name, file_extension="msh"
                )
                reference_file = self.get_path_to_reference_file(
                    data_model=data_model, model_name=model_name, file_extension="msh"
                )
                self.compare_mesh_qualities(mesh_file, reference_file)

                # Compare the regions files:
                regions_file = self.get_path_to_generated_file(
                    data_model=data_model,
                    model_name=model_name,
                    file_extension="regions",
                )
                reference_regions_file = self.get_path_to_reference_file(
                    data_model=data_model,
                    model_name=model_name,
                    file_extension="regions",
                )
                print('Comparing'
                      f'{regions_file}'
                      'with'
                      f'{reference_regions_file}'
                      )
                self.compare_json_or_yaml_files(regions_file, reference_regions_file)

    def test_ConductorAC_Rutherford(self):
        """
        Checks if mesh generators work correctly by comparing the number
        of entities in the generated geometry file to the reference file that was
        checked manually.
        """
        model_names = [
            "TEST_CAC_Rutherford",
        ]

        for model_name in model_names:
            with self.subTest(model_name=model_name):
                data_model = self.get_data_model(model_name)

                self.generate_mesh(data_model, model_name)

                # Compare the number of entities with the reference file:
                mesh_file = self.get_path_to_generated_file(
                    data_model=data_model, model_name=model_name, file_extension="msh"
                )
                reference_file = self.get_path_to_reference_file(
                    data_model=data_model, model_name=model_name, file_extension="msh"
                )
                self.compare_mesh_qualities(mesh_file, reference_file)

                # Compare the regions files:
                regions_file = self.get_path_to_generated_file(
                    data_model=data_model,
                    model_name=model_name,
                    file_extension="regions",
                )
                reference_regions_file = self.get_path_to_reference_file(
                    data_model=data_model,
                    model_name=model_name,
                    file_extension="regions",
                )
                self.compare_json_or_yaml_files(regions_file, reference_regions_file)

    def test_ConductorAC_CC(self):
        """
        Checks if mesh generators work correctly by comparing the number
        of entities in the generated geometry file to the reference file that was
        checked manually.
        """
        model_names = [
            "TEST_CAC_CC",
        ]

        for model_name in model_names:
            with self.subTest(model_name=model_name):
                data_model = self.get_data_model(model_name)

                self.generate_mesh(data_model, model_name)

                # Compare the number of entities with the reference file:
                mesh_file = self.get_path_to_generated_file(
                    data_model=data_model, model_name=model_name, file_extension="msh"
                )
                reference_file = self.get_path_to_reference_file(
                    data_model=data_model, model_name=model_name, file_extension="msh"
                )
                self.compare_mesh_qualities(mesh_file, reference_file)

                # Compare the regions files:
                regions_file = self.get_path_to_generated_file(
                    data_model=data_model,
                    model_name=model_name,
                    file_extension="regions",
                )
                reference_regions_file = self.get_path_to_reference_file(
                    data_model=data_model,
                    model_name=model_name,
                    file_extension="regions",
                )
                self.compare_json_or_yaml_files(regions_file, reference_regions_file)

    # def test_CCT(self):
    #     """
    #     Checks if CCT geometry generators work correctly by comparing the number
    #     of entities in the generated geometry file to the reference file that was
    #     checked manually.
    #     """
    #     model_names = [
    #         "MCBRD_2d2a_2n2a_0i",
    #     ]

    #     for model_name in model_names:
    #         with self.subTest(model_name=model_name):
    #             data_model = self.get_data_model(model_name)

    #             # data_model can be modified here if necessary
    #             # Example:

    #             # data_model.magnet.mesh.wi.axne = 30

    #             self.generate_mesh(data_model, model_name)

    #             # Compare the number of entities with the reference file:
    #             mesh_file = self.get_path_to_generated_file(
    #                 data_model=data_model, model_name=model_name, file_extension="msh"
    #             )
    #             reference_file = self.get_path_to_reference_file(
    #                 data_model=data_model, model_name=model_name, file_extension="msh"
    #             )
    #             self.compare_mesh_qualities(mesh_file, reference_file)


    def test_Multipole(self):
        """
        Checks if Multipole mesh generators work correctly by comparing against
        reference files that were checked manually.
        """
        model_names = [
            "TEST_MULTIPOLE_MBH_1in1_TSA_withQH",
            "TEST_MULTIPOLE_MBH_1in1_TSA",
            "TEST_MULTIPOLE_MBH_1in1_REF",
            "TEST_MULTIPOLE_SMC_TSA_withQH",
            "TEST_MULTIPOLE_SMC_TSA",
            "TEST_MULTIPOLE_SMC_REF",
            "TEST_MULTIPOLE_4COND_TSA",
            "TEST_MULTIPOLE_FALCOND_C_TSA_COLLAR_POLE"
        ]

        for model_name in model_names:
            with self.subTest(model_name=model_name):
                data_model = self.get_data_model(model_name)

                is_TSA = "TSA" in model_name

                self.generate_mesh(data_model, model_name)

                thermal_model_name = f"{model_name}_TH"
                electromagnetic_model_name = f"{model_name}_EM"

                # Compare mesh qualities with the reference file:
                mesh_file_EM = self.get_path_to_generated_file(
                    data_model=data_model, model_name=electromagnetic_model_name, file_extension="msh"
                )
                reference_file_EM = self.get_path_to_reference_file(
                    data_model=data_model, model_name=electromagnetic_model_name, file_extension="msh"
                )
                self.compare_mesh_qualities(mesh_file_EM, reference_file_EM)

                mesh_file_TH = self.get_path_to_generated_file(
                    data_model=data_model, model_name=thermal_model_name, file_extension="msh"
                )
                reference_file_TH = self.get_path_to_reference_file(
                    data_model=data_model, model_name=thermal_model_name, file_extension="msh"
                )
                self.compare_mesh_qualities(mesh_file_TH, reference_file_TH)

                # Compare the reg files:
                regions_file_EM = self.get_path_to_generated_file(
                    data_model=data_model,
                    model_name=electromagnetic_model_name,
                    file_extension="reg",
                )
                reference_regions_file_EM = self.get_path_to_reference_file(
                    data_model=data_model,
                    model_name=electromagnetic_model_name,
                    file_extension="reg",
                )
                self.compare_json_or_yaml_files(regions_file_EM, reference_regions_file_EM)

                regions_file_TH = self.get_path_to_generated_file(
                    data_model=data_model,
                    model_name=thermal_model_name,
                    file_extension="reg",
                )
                reference_regions_file_TH = self.get_path_to_reference_file(
                    data_model=data_model,
                    model_name=thermal_model_name,
                    file_extension="reg",
                )
                self.compare_json_or_yaml_files(regions_file_TH, reference_regions_file_TH)

                # Compare the aux files:
                aux_file_EM = self.get_path_to_generated_file(
                    data_model=data_model,
                    model_name=electromagnetic_model_name,
                    file_extension="aux",
                )
                reference_aux_file_EM = self.get_path_to_reference_file(
                    data_model=data_model,
                    model_name=electromagnetic_model_name,
                    file_extension="aux",
                )
                self.compare_json_or_yaml_files(aux_file_EM, reference_aux_file_EM)

                aux_file_TH = self.get_path_to_generated_file(
                    data_model=data_model,
                    model_name=thermal_model_name,
                    file_extension="aux",
                )
                reference_aux_file_TH = self.get_path_to_reference_file(
                    data_model=data_model,
                    model_name=thermal_model_name,
                    file_extension="aux",
                )
                self.compare_json_or_yaml_files(aux_file_TH, reference_aux_file_TH)

                if is_TSA:
                    # Compare the reco file:
                    reco_file = self.get_path_to_generated_file(
                        data_model=data_model,
                        model_name=thermal_model_name,
                        file_extension="reco",
                    )
                    reference_reco_file = self.get_path_to_reference_file(
                        data_model=data_model,
                        model_name=thermal_model_name,
                        file_extension="reco",
                    )
                    self.compare_json_or_yaml_files(reco_file, reference_reco_file)

                # Compare the mesh.yaml file:
                mesh_yaml_file = self.get_path_to_generated_file(
                    data_model=data_model,
                    model_name="mesh",
                    file_extension="yaml",
                )
                reference_mesh_yaml_file = self.get_path_to_reference_file(
                    data_model=data_model,
                    model_name="mesh",
                    file_extension="yaml",
                )
                self.compare_json_or_yaml_files(mesh_yaml_file, reference_mesh_yaml_file)


if __name__ == "__main__":
    unittest.main()



import os, logging
import numpy as np
import matplotlib.pyplot as plt

logger = logging.getLogger('FiQuS')

class PostProcess:
    def __init__(self, solution_folder_path):
        self.solution_folder_path = solution_folder_path

        self.plot_transport()
        # self.plot_strand_split()
        self.plot_fluxhyst()
        self.plot_power_rohf()

    def load_standard_txt(self, file_rel_path, skiprows=1):
        """Load the content of a txt file into a nested numpy array

        :param file_rel_path: relative file path in the solution folder (with file extension)
        :param skiprows: number of rows to skip at the top of the txt file, defaults to 1
        :return: nested np.array with time at index 0
        """
        file_path = os.path.join(self.solution_folder_path, file_rel_path)
        return np.loadtxt(file_path, skiprows=skiprows).T if os.path.isfile(file_path) else None

    def plot_transport(self):
        """ Plots transport current and voltage over time if existing """

        Vt_data = self.load_standard_txt('txt_files\\Vt.txt')
        It_data = self.load_standard_txt('txt_files\\It.txt') 

        if all(x is not None for x in (It_data, Vt_data)):
            fig, ax1 = plt.subplots()
            ax1.set_title("Transport current and voltage")
            ax1.set_xlabel(r'Time $(s)$')
            ax1.set_ylabel(r'Current $(A)$', color = 'red')
            ax1.plot(It_data[0], It_data[1], color = 'red')
            ax1.grid(axis='x', linestyle='dotted')
            ax1.tick_params(axis='y', labelcolor = 'red')
            
            ax2 = ax1.twinx()
            ax2.set_ylabel(r'Voltage $(V)$', color = 'blue')
            ax2.plot(Vt_data[0], Vt_data[1], color = 'blue')
            ax2.tick_params(axis='y', labelcolor = 'blue')
            ax2.grid(axis='y', linestyle='dotted')
            fig.tight_layout()

    def plot_fluxhyst(self):
        """ Plots the internal flux hysteresis and resulting voltages if existing (ROHF enabled and stranded strands) """

        fluxhyst_data = self.load_standard_txt('txt_files\\flux_hyst.txt')
        voltage_rohf = self.load_standard_txt('txt_files\\voltage_rohf.txt')
        Is_data = self.load_standard_txt('txt_files\\Is.txt')  

        if all(x is not None for x in (fluxhyst_data, Is_data)):
            # plot only the first quarter of the strands because of symmetry ...
            q_idx = round((len(fluxhyst_data)-1) / 4)
            colors = plt.cm.rainbow(np.linspace(0,1,q_idx))
            plt.figure(figsize=(10,5))
            # Flux in strand currents
            plt.subplot(121)
            plt.title('ROHF Flux hysteresis')
            for i in range(1,q_idx+1):
                plt.plot(Is_data[i], fluxhyst_data[i], label='Strand '+str(i), color=colors[i-1]) 
            plt.xlabel(r"Strand current (A)")
            plt.ylabel(r'Internal flux (Wb/m$)')
            plt.legend()  

            # Resulting strand ROHF voltages
            plt.subplot(122)
            plt.title('ROHF voltages')
            if voltage_rohf is not None:
                for i in range(1, q_idx+1):
                    plt.plot(voltage_rohf[0], voltage_rohf[i], label='Strand '+str(i), color=colors[i-1])
                    #plt.plot(self.linflux_voltage[0], self.linflux_voltage[i], linestyle='dashed', color=colors[i-1])
            plt.xlabel('Time (s)')
            plt.ylabel(r'Voltage (V/m)')
            plt.legend()

    def plot_strand_split(self):
        """ Plots the Strand currents and voltages for stranded strands enabled """

        Vt_data = self.load_standard_txt('txt_files\\Vt.txt')
        It_data = self.load_standard_txt('txt_files\\It.txt') 
        Vs_data = self.load_standard_txt('txt_files\\Vs.txt')
        Is_data = self.load_standard_txt('txt_files\\Is.txt')  

        if all(x is not None for x in (It_data, Vt_data, Is_data, Vs_data)):
            # plot only the first quarter of the strands because of symmetry ...
            q_idx = round((len(Vs_data)-1) / 4)
            colors = plt.cm.rainbow(np.linspace(0,1,q_idx))

            plt.figure(figsize=(9,5))
            plt.subplot(121)
            plt.title('Strand voltages')
            plt.plot(Vt_data[0], Vt_data[1], label='Total transport voltage')
            for i in range(1,q_idx+1):
                plt.plot(Vs_data[0], Vs_data[i], label='Strand '+str(i), color=colors[i-1])  
            plt.plot(Vs_data[0], sum(Vs_data[1:]), 'r.--', label=r'$\sum \ V_s$')
            #plt.scatter(self.Vs_data[0], self.Vs_data[36], marker='x', label='Strand 36')
            plt.xlabel(r"Time (s)")
            plt.ylabel(r'Voltage (V/m)')
            plt.legend()       

            plt.subplot(122)
            plt.title('Strand currents')
            plt.plot(It_data[0], It_data[1], label='Total transport current')
            for i in range(1,q_idx+1):
                plt.plot(Is_data[0], Is_data[i], label='Strand '+str(i), color=colors[i-1])
            plt.plot(Is_data[0], sum(Is_data[1:]), 'r.--', label=r'$\sum \ I_s$')  
            #plt.scatter(self.Is_data[0], self.Is_data[36], marker='x', label='Strand 36')
            plt.xlabel(r"Time (s)")
            plt.ylabel(r'Current (A)')       
            plt.legend()  

    def plot_power_rohf(self):
        """ Plots the ROHF related power contributions if existing """

        power = self.load_standard_txt('txt_files\\power.txt')
        power_hyst_rohf = self.load_standard_txt('txt_files\\power_hyst_ROHF.txt')
        power_eddy_rohf = self.load_standard_txt('txt_files\\power_eddy_ROHF.txt')

        if power is not None and len(power) == 5:
            colors = plt.cm.rainbow(np.linspace(0,1,4))

            plt.figure()
            plt.title('Instantaneous power loss contributions')
            #plt.plot(self.power[0], self.power[1], label='Total', color=colors[0])   
            plt.plot(power[0], power[2], label='hyst powerfile', linestyle='--', color=colors[1])
            plt.plot(power_hyst_rohf[0], sum(power_hyst_rohf[1:]), label='Hyst', color=colors[1])
            plt.plot(power[0], power[3], label='Eddy powerfile', linestyle='--', color=colors[2])
            plt.plot(power_eddy_rohf[0], sum(power_eddy_rohf[1:]), label='Eddy', color=colors[2]) 
            #plt.plot(power_hyst_strands[0], sum(power_hyst_strands[1:]), label='strands sum')
            plt.xlabel(r"Time (s)")
            plt.ylabel(r'Power loss (W/m)') 
            plt.legend()
        
    def show(self):
        """ Display all generated plots at once """
        plt.show() if len(plt.get_fignums()) > 0 else logger.info('- NO POSTPROC DATA -')





from pydantic import BaseModel
from typing import List, Dict, Union, Optional, Literal


class Region(BaseModel):
    name: Optional[str] = None
    number: Optional[int] = None


class Regions(BaseModel):
    names: Optional[List[str]] = None
    numbers: Optional[List[int]] = None


class Regions2(BaseModel):
    names: Optional[List[List[str]]] = None
    numbers: Optional[List[List[int]]] = None


class TwoParBoundaryRegions(Regions2):
    values: Optional[List[List[Union[float, str]]]] = None


class TwoParBoundaryRegions2(Regions):
    values: Optional[List[Union[float, str]]] = None


class OneParBoundaryRegions(Regions2):
    value: Optional[List[float]] = None

class InducedRegions(Regions):
    sigmas: Optional[List[float]] = None
    mu_rs: Optional[List[float]] = None


class PoweredRegions(InducedRegions):
    currents: Optional[List[float]] = None


class InsulatorRegions(InducedRegions):
    pass # duplicate class


class IronRegions(InducedRegions):
    pass # duplicate class


class AirRegion(Region):
    sigma: Optional[float] = None
    mu_r: Optional[float] = None


class AirFarFieldRegions(Regions):
    radius_in: Optional[float] = None
    radius_out: Optional[float] = None


class NonPermeableSourceRegion(AirRegion):
    pass # duplicate class


class SourceFreeRegion(AirRegion):
    pass # duplicate class


class Iron(BaseModel):
    vol: IronRegions = IronRegions()  # volume region
    surf: Regions = Regions()  # surface region
    curves: Dict[str, List[str]] = {}  # curves region


class Induced(BaseModel):
    vol: InducedRegions = InducedRegions()  # volume region
    surf_th: Regions = Regions()  # surface region
    surf_in: Regions = Regions()  # input terminal surface region
    surf_out: Regions = Regions()  # output terminal surface region
    cochain: Regions = Regions()  # winding cochain (cut)

class Insulator(BaseModel):
    vol: InsulatorRegions = InsulatorRegions()  # volume region
    surf: Regions = Regions()  # surface region
    curve: Regions = Regions()  # curve region

class Powered(Induced, Insulator):
    vol: PoweredRegions = PoweredRegions()  # volume region
    vol_in: Region = Region()  # input terminal volume region
    vol_out: Region = Region()  # output terminal volume region
    conductors: Dict[str, List[str]] = {}  # conductor types
    surf_insul: Regions = Regions()  # turn-to-turn insulation surfaces


class Air(BaseModel):
    vol: AirRegion = AirRegion()  # volume region
    surf: Region = Region()  # surface region
    line: Region = Region()  # line region
    point: Regions = Regions()  # point region
    cochain: Regions = Regions()  # air cochain (cut)


class AirFarField(BaseModel):
    vol: AirFarFieldRegions = AirFarFieldRegions()  # volume region
    surf: Region = Region()  # surface region


class NonPermeableSource(BaseModel):
    vol: NonPermeableSourceRegion = NonPermeableSourceRegion()  # volume region
    surf: Region = Region()  # surface region


class SourceFree(BaseModel):
    vol: SourceFreeRegion = SourceFreeRegion()  # volume region
    surf: Region = Region()  # surface region

class RobinCondition_collar(BaseModel):
    bc: TwoParBoundaryRegions2 = TwoParBoundaryRegions2()

class RobinCondition_grouped(BaseModel):
    bc: TwoParBoundaryRegions = TwoParBoundaryRegions()
    groups: Dict[str, List[int]] = {}

class NeumannCondition_grouped(BaseModel):
    bc: OneParBoundaryRegions = OneParBoundaryRegions()
    groups: Dict[str, List[int]] = {}

class DirichletCondition_grouped(BaseModel):
    bc: OneParBoundaryRegions = OneParBoundaryRegions()
    groups: Dict[str, List[int]] = {}


class ThermalBoundaryConditions(BaseModel):
    temperature: DirichletCondition_grouped = DirichletCondition_grouped()
    heat_flux: NeumannCondition_grouped = NeumannCondition_grouped()
    cooling: RobinCondition_grouped = RobinCondition_grouped()
    collar: RobinCondition_collar = RobinCondition_collar()


class SymmetryBoundaryConditions(BaseModel):
    normal_free: Region = Region()
    tangential_free: Region = Region()


class BoundaryConditions(BaseModel):
    thermal: ThermalBoundaryConditions = ThermalBoundaryConditions()
    symmetry: SymmetryBoundaryConditions = SymmetryBoundaryConditions()


class InsulationType(BaseModel):
    layers_number: List[int] = []
    thin_shells: List[List[int]] = []
    layers_material: List[List[str]] = []
    thicknesses: List[List[float]] = []
    correction_factors: List[float] = [] # same correction factor per group. Can be used to scale the thin shell length
    label: List[List[Union[int, None]]] = (
        []
    )  # useful to indicate which quench heater a SS element refers to


class ThinShell(BaseModel):
    groups: Dict[str, List[int]] = {} # -> Checkboard convention of the hts
    mid_turns_layers_poles: Optional[List[int]] = None # indices of midlayers between HT, layers and poles
    ts_collar_groups: Dict[str, List[int]] = {} # groups of mid collar thin shells -> Checkboard convention
    ts_pole_groups: Dict[str, List[int]] = {} # groups of mid collar thin shells -> Checkboard convention

    second_group_is_next: Dict[str, List[int]] = {}
    normals_directed: Dict[str, List[int]] = {}
    bdry_curves: Dict[Literal["collar", "poles", "outer_collar"], List[int]] = {} # save the boundary curves of a specific region, used in TSA

    # insulation types, only mid layers
    insulation_types: InsulationType = InsulationType()
    quench_heaters: InsulationType = InsulationType()
    collar: InsulationType = InsulationType()
    poles: InsulationType = InsulationType()

class PostProc(BaseModel):
    vol: Regions = Regions()  # postprocessing volumes general
    surf: Regions = Regions()  # postprocessing volumes general
    line: Regions = Regions()  # postprocessing volumes general
    point: Regions = Regions()  # postprocessing volumes general


class RegionsModel(BaseModel):
    powered: Dict[str, Powered] = {} #coils
    induced: Dict[str, Induced] = {} #wedges
    insulator: Insulator = Insulator()
    iron_yoke: Iron = Iron() # iron yoke
    collar: Iron = Iron() # collar is a type of iron region
    ref_mesh: Iron = Iron()  # reference mesh
    poles: Iron = Iron()  # poles

    air: Air = Air()
    air_far_field: AirFarField = AirFarField()
    thin_shells: ThinShell = ThinShell() # Includes collar thin shells
    projection_points: Region = Region()
    boundaries: BoundaryConditions = BoundaryConditions()
    postproc_th: PostProc = PostProc()
    postproc_em: PostProc = PostProc()




import os
import math
import csv
import re
import logging

import gmsh
import json
import numpy as np
import pandas as pd
from pathlib import Path
import matplotlib.pyplot as plt
from matplotlib import cm
from mpl_toolkits.mplot3d.art3d import Poly3DCollection

from fiqus.utils.Utils import GmshUtils
from fiqus.parsers.ParserCOND import ParserCOND
from fiqus.utils.Utils import FilesAndFolders as uff
from fiqus.data.RegionsModelFiQuS import RegionsModel

logger = logging.getLogger('FiQuS')

class Post_Process:
    """
    Postprocessing class for CWS
    """
    def __init__(self, fdm, verbose=True):
        """
        Class to cct models postprocessing
        :param fdm: FiQuS data model
        :param verbose: If True more information is printed in python console.
        """
        self.current_FiQuS = {}
        self.fdm = fdm
        self.model_folder = os.path.join(os.getcwd())

        self.mesh_folder = Path(self.model_folder).parent
        self.geom_folder = Path(self.mesh_folder).parent
        self.magnet_name = fdm.general.magnet_name
        self.verbose = verbose

    def cleanup(self):
        """
            This function is used to remove .msh, .pre and .res files from the solution folder, as they may be large and not needed.
        """
        magnet_name = self.fdm.general.magnet_name
        cleanup = self.fdm.magnet.postproc.cleanup

        if cleanup.remove_res_file:
            res_file_path = os.path.join(self.model_folder, f"{magnet_name}.res")
            if os.path.exists(res_file_path):
                os.remove(res_file_path)
                logger.info(f"Removed {magnet_name}.res")

        if cleanup.remove_pre_file:
            pre_file_path = os.path.join(self.model_folder, f"{magnet_name}.pre")
            if os.path.exists(pre_file_path):
                os.remove(pre_file_path)
                logger.info(f"Removed {magnet_name}.pre")

        if cleanup.remove_msh_file:
            msh_file_path = os.path.join(self.mesh_folder, f"{magnet_name}.msh")
            if os.path.exists(msh_file_path):
                os.remove(msh_file_path)
                logger.info(f"Removed {magnet_name}.msh")
